/*
 * HP HIL Protocol linux-input device driver for keyboard devices
 *
 * Copyright (c) 2001-2004 Brian S. Julin
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL").
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *
 * References:
 * HP-HIL Technical Reference Manual.  Hewlett Packard Product No. 45918A
 *
 */

#include <linux/hil.h>
#include <linux/input.h>
#include <linux/serio.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/pci_ids.h>


/* Module administrativa */
#define PREFIX "HIL KBD: "
#define HIL_GENERIC_NAME "HIL keyboard"
MODULE_AUTHOR("Brian S. Julin <bri@calyx.com>");
MODULE_DESCRIPTION(HIL_GENERIC_NAME " driver");
MODULE_LICENSE("Dual BSD/GPL");


/* Import keycode translation tables from header. */
#define HIL_KBD_SET1_UPBIT 0x01
#define HIL_KBD_SET1_SHIFT 1
static unsigned int hil_kbd_set1[HIL_KEYCODES_SET1_TBLSIZE] = 
	{ HIL_KEYCODES_SET1 };
#define HIL_KBD_SET2_UPBIT 0x01
#define HIL_KBD_SET2_SHIFT 1
	/* Set2 is "user defined." */
#define HIL_KBD_SET3_UPBIT 0x80
#define HIL_KBD_SET3_SHIFT 0
static unsigned int hil_kbd_set3[HIL_KEYCODES_SET3_TBLSIZE] =
	{ HIL_KEYCODES_SET3 };


/* Import HIL locale to locale strings mapping from header. */
static char hil_language[][16] = { HIL_LOCALE_MAP };


/* Per-keyboard structure definition. */
struct hil_kbd {
	/* Ties into the linux-input system */
	struct input_dev dev;
	struct serio *serio;
	char devphys[11];

	/* Items required to cascade a subordinate driver for pointers */
	struct serio cascade;
	char cascphys[14];
	int casc_need_reg, axdepth, casc_wshift;
	hil_frame cascbuf;

	int ape; /* When set we use only read operations (APE compat mode.) */

	/* Input buffer and index for packets from HIL bus. */
	hil_frame data[16];
	int idx4; /* four counts per frame */

	/* Raw device info records from HIL bus, see hil.h for fields. */
	uint8_t	idd[16];	/* DID byte and IDD record */
	uint8_t rsc[16];	/* RSC record */
	uint8_t	exd[16];	/* EXD record */
	uint8_t	rnm[17];	/* RNM record + NULL term. */

	/* Something to sleep around with. */
	struct semaphore sem;
};


/* Until the input/serio core gets ported to the new driver model, we will 
 * need to free our structures *after* unregistering our driver, due to 
 * the cascaded serios, and register our cascades later as well.  Hopefully
 * these hacks can go away soon.
 */
static struct hil_kbd *hil_kbd_freeme[32] = { NULL, };
static unsigned long hil_kbd_idsused[1];


/* Convenience functions -- send a hil_frame via serio char-by-char io. */

static int hil_intr_frame(struct serio *s, struct serio_dev *d, hil_frame f) {
	if (d) {
		d->interrupt(s, (f >> 24) & 0xff, 0, NULL);
		d->interrupt(s, (f >> 16) & 0xff, 0, NULL);
		d->interrupt(s, (f >> 8)  & 0xff, 0, NULL);
		d->interrupt(s, (f >> 0)  & 0xff, 0, NULL);
		return 0;
	}
	return -EIO;
}

static void hil_write_frame(struct serio *s, hil_frame f) {
	s->write(s, (f >> 24) & 0xff);
	s->write(s, (f >> 16) & 0xff);
	s->write(s, (f >> 8)  & 0xff);
	s->write(s, (f >> 0)  & 0xff);
}


/* Functions dealing with cascaded pointer devices.  In APE mode even
 * standalone mice run through the hil_kbd cascade.   In either this case,
 * or with genuine keyboard-embedded pointers, when in APE mode we have 
 * no IDD, so we have to autodetect characteristics.
 */

/* Register another serio for a cascaded mouse driver on a combo device */
static void hil_kbd_casc_reg(struct hil_kbd *kbd) {
	int axnum, cnt;

	if (!kbd->ape) { /* MLC will have provided a correct idd record. */
		kbd->axdepth = (kbd->idd[1] & HIL_IDD_HEADER_16BIT) ? 2 : 1;
		goto done;
	}

	/* No genuine IDD, so examine the record. */

	/* Temp hack: assume a relative mouse, max out number of buttons. */
	kbd->idd[0] = HIL_IDD_DID_TYPE_REL;
	kbd->idd[2] = 255;
	kbd->idd[3] = 0;
	kbd->idd[4] = HIL_IDD_IOD_NBUTTON_MASK & 7;
	kbd->idd[5] = HIL_CMD_IDD;

	if (kbd->idd[1]) goto absrel; /* We have already figured axis size */

	axnum = kbd->data[0] & HIL_POL_NUM_AXES_MASK;
	if (!axnum) goto defer;
	
	cnt = kbd->idx4/4 - 2;
	if (cnt < 1) goto defer;

	/* Iff no key data is indicated the length is directly calculated. */
	if ((kbd->data[0] & HIL_POL_CHARTYPE_MASK) == HIL_POL_CHARTYPE_NONE) {
		switch (cnt / axnum) {
		case 2:
			kbd->idd[1] |= HIL_IDD_HEADER_16BIT;
		case 1:
			kbd->idd[1] |= HIL_IDD_HEADER_IOD; 
			kbd->idd[1] |= HIL_IDD_HEADER_AXSET_MASK;
			kbd->axdepth = cnt/axnum;
			goto absrel;
		default:
			goto defer;
		}
	}
	/* Otherwise we can sometimes tell the depth is 1 if the 
	 * packet is too small to contain 2-byte data for all listed axis.
	 */
	if (cnt < axnum * 2) {
		kbd->idd[1] |= HIL_IDD_HEADER_IOD; 
		kbd->idd[1] |= HIL_IDD_HEADER_AXSET_MASK & 2;
		kbd->axdepth = 1;
		goto absrel;
	}
	/* We still do not know and must wait for more data. */
	goto defer;
 absrel:
	if (kbd->idd[0]) goto done; /* We have already determined this. */

	/* TODO: This is trickier.  We have to keep track of a lot of values 
	 * and guess whether we are being sent absolute or relative values.
	 *
	 * Maybe there is code to be ripped from other drivers for this...
	 */
 defer:
	kbd->idx4 = 0;
	up(&kbd->sem);
	return;
 done:
	printk(KERN_INFO "serio: %s: %s.\n", kbd->cascphys, kbd->cascade.name);
	kbd->casc_need_reg = 2; /* fall-through and process this packet */
	serio_register_port(&kbd->cascade);
	return;
}

/* Forward a pointer button to a cascaded pointer driver. */
static void hil_kbd_fwd_ptr_btn(struct hil_kbd *kbd, unsigned int key) {
	struct serio *s;
	struct serio_dev *d;

	if (kbd->casc_need_reg != 2) return;
	s = &kbd->cascade;
	d = s->dev;
	if (!d) return;

	hil_intr_frame(s, d, 
		       HIL_ERR_INT | /*devid<<8 |*/ HIL_POL_CHARTYPE_SET1); 
	hil_intr_frame(s, d, HIL_ERR_INT | /*devid<<8 |*/ (key & 0xff)); 
	hil_intr_frame(s, d, HIL_ERR_INT | HIL_FRAME_CMD |
		       /* devid >> 8 | */ HIL_CMD_POL | 2);
}

/* Forward a pointer axis to a cascaded pointer driver. */
static int hil_kbd_fwd_ptr_ax(struct hil_kbd *kbd) {
	int cnt, axnum;
	axnum = kbd->data[0] & HIL_POL_NUM_AXES_MASK;
	struct serio *s;
	struct serio_dev *d;
	hil_frame data;

	if (kbd->casc_need_reg != 2) goto out;
	s = &kbd->cascade;
	d = s->dev;
	if (!d) goto out;

	for (cnt = 0; cnt <= axnum * kbd->axdepth; cnt++) {
		/* Send the axes data out to the mouse driver. */
		data = kbd->data[cnt];
		if (!cnt) data &= ~HIL_POL_CHARTYPE_MASK;
		hil_intr_frame(s, d, data);
	}
	hil_intr_frame(s, d, HIL_ERR_INT | HIL_FRAME_CMD | HIL_CMD_POL | cnt);
 out:
	return (axnum * kbd->axdepth + 1);
}


/* Functions to buffer/process data from the HIL */

static void hil_kbd_process_record(struct hil_kbd *kbd)
{
	struct input_dev *dev = &kbd->dev;
	hil_frame *data = kbd->data;
	hil_frame p;
	int idx, i, cnt;

	idx = kbd->idx4/4;
	p = data[idx - 1];

	if ((p & ~(HIL_CMDCT_POL | HIL_FRAME_ADDR_MASK)) == 
	    (HIL_ERR_INT | HIL_FRAME_CMD | HIL_CMD_POL)) goto report;
	if ((p & ~(HIL_CMDCT_RPL | HIL_FRAME_ADDR_MASK)) == 
	    (HIL_ERR_INT | HIL_FRAME_CMD | HIL_CMD_RPL)) goto report;

	/* Not a poll response.  See if we are loading config records. */
	switch (p & HIL_FRAME_DATA_MASK) {
	case HIL_CMD_IDD:
		for (i = 0; i < idx; i++)
			kbd->idd[i] = kbd->data[i] & HIL_FRAME_DATA_MASK;
		for (; i < 16; i++)
			kbd->idd[i] = 0;
		break;
	case HIL_CMD_RSC:
		for (i = 0; i < idx; i++)
			kbd->rsc[i] = kbd->data[i] & HIL_FRAME_DATA_MASK;
		for (; i < 16; i++)
			kbd->rsc[i] = 0;
		break;
	case HIL_CMD_EXD:
		for (i = 0; i < idx; i++)
			kbd->exd[i] = kbd->data[i] & HIL_FRAME_DATA_MASK;
		for (; i < 16; i++)
			kbd->exd[i] = 0;
		break;
	case HIL_CMD_RNM:
		for (i = 0; i < idx; i++)
			kbd->rnm[i] = kbd->data[i] & HIL_FRAME_DATA_MASK;
		for (; i < 17; i++)
			kbd->rnm[i] = '\0';
		break;
	default:
		/* These occur when device isn't present */
		if (p == (HIL_ERR_INT | HIL_FRAME_CMD)) break; 
		/* Anything else we'd like to know about. */
		printk(KERN_WARNING PREFIX "Got unknown record (%x)\n", p);
		break;
	}
	goto out;

 report:
	if (kbd->casc_need_reg == 1) hil_kbd_casc_reg(kbd);

	cnt = 1;
	if (kbd->data[0] & HIL_POL_NUM_AXES_MASK) 
		cnt = hil_kbd_fwd_ptr_ax(kbd);

	switch (kbd->data[0] & HIL_POL_CHARTYPE_MASK) {
	case HIL_POL_CHARTYPE_NONE:
		break;
	case HIL_POL_CHARTYPE_ASCII:
		while (cnt < idx - 1)
			input_report_key(dev, kbd->data[cnt++] & 0x7f, 1);
		break;
	case HIL_POL_CHARTYPE_RSVD1:
	case HIL_POL_CHARTYPE_RSVD2:
	case HIL_POL_CHARTYPE_BINARY:
		while (cnt < idx - 1)
			input_report_key(dev, kbd->data[cnt++], 1);
		break;
	case HIL_POL_CHARTYPE_SET1:
		while (cnt < idx - 1) {
			unsigned int key;
			int up;
			key = kbd->data[cnt++];
			if (((key & 0xff) >= 0x80) && 
			    ((key & 0xff) <= 0x8f) && (kbd->axdepth)) {
				hil_kbd_fwd_ptr_btn(kbd, key);
			}
			else {
				up = key & HIL_KBD_SET1_UPBIT;
				key &= (~HIL_KBD_SET1_UPBIT & 0xff);
				key = hil_kbd_set1[key >> HIL_KBD_SET1_SHIFT];
				if (key != KEY_RESERVED)
				  input_report_key(dev, key, !up);
			}
		}
		break;
	case HIL_POL_CHARTYPE_SET2:
		while (cnt < idx - 1) {
			unsigned int key;
			int up;
			key = kbd->data[cnt++];
			up = key & HIL_KBD_SET2_UPBIT;
			key &= (~HIL_KBD_SET1_UPBIT & 0xff);
			key = key >> HIL_KBD_SET2_SHIFT;
			if (key != KEY_RESERVED)
				input_report_key(dev, key, !up);
		}
		break;
	case HIL_POL_CHARTYPE_SET3:
		while (cnt < idx - 1) {
			unsigned int key;
			int up;
			key = kbd->data[cnt++];
			up = key & HIL_KBD_SET3_UPBIT;
			key &= (~HIL_KBD_SET1_UPBIT & 0xff);
			key = hil_kbd_set3[key >> HIL_KBD_SET3_SHIFT];
			if (key != KEY_RESERVED)
				input_report_key(dev, key, !up);
		}
		break;
	}
 out:
	kbd->idx4 = 0;
	up(&kbd->sem);
}

static void hil_kbd_process_err(struct hil_kbd *kbd) {
	printk(KERN_WARNING PREFIX "errored HIL packet\n");
	kbd->idx4 = 0;
	up(&kbd->sem);
}

static irqreturn_t hil_kbd_interrupt(struct serio *serio, unsigned char data, 
				     unsigned int flags, struct pt_regs *regs)
{
	struct hil_kbd *kbd;
	hil_frame packet;
	int idx;

	kbd = (struct hil_kbd *)serio->private;
	if (kbd == NULL) {
		BUG();
		return IRQ_HANDLED;
	}

	/* Records cannot be over 16 frames long. */
	if (kbd->idx4 >= (16 * sizeof(hil_frame))) {
		hil_kbd_process_err(kbd);
		return IRQ_HANDLED;
	}

	/* Assemble 1-byte data into buffer of 4-byte hil_frames. */
	idx = kbd->idx4/4;
	if (!(kbd->idx4 % 4)) kbd->data[idx] = 0;
	packet = kbd->data[idx];
	packet |= ((hil_frame)data) << ((3 - (kbd->idx4 % 4)) * 8);
	kbd->data[idx] = packet;
	if ((++(kbd->idx4)) % 4) return IRQ_HANDLED;

	/* All normal hil_frames have a normal "interrupt" status code */
	if ((packet & 0xffff0000) != HIL_ERR_INT) {
		hil_kbd_process_err(kbd);
		return IRQ_HANDLED;
	}

	/* Records of N hil_frames will terminate with a command. */
	if (packet & HIL_FRAME_CMD) hil_kbd_process_record(kbd);

	return IRQ_HANDLED;
}


/* Linkage for cascaded pointer sub-devices into the linux-input system. */

static int hil_kbd_cascade_write(struct serio *port, unsigned char c) {
	struct hil_kbd *kbd;
	uint8_t *idx, *last;
	struct serio_dev *dev;

	kbd = container_of(port, struct hil_kbd, cascade);

	/* Group consecutive 1-byte writes into a 4-byte hil_frame */
	kbd->cascbuf |= ((hil_frame)c) << kbd->casc_wshift;
	if (!kbd->casc_wshift) goto process;
	kbd->casc_wshift -= 8;
	return 0;

 process:
	/* We only emulate a few 1-frame commands. */
	if (!(kbd->cascbuf & HIL_FRAME_CMD)) goto iserr;

	switch (kbd->cascbuf & HIL_FRAME_DATA_MASK) {
	case HIL_CMD_IDD:
		idx = kbd->idd;
		break;
	case HIL_CMD_RSC:
		idx = kbd->rsc;
		break;
	case HIL_CMD_EXD:
		idx = kbd->exd;
		break;
	case HIL_CMD_RNM:
		idx = kbd->rnm;
		break;
	default:
		goto iserr;
	}
	kbd->casc_wshift = 24;
	kbd->cascbuf = 0;

	dev = port->dev;
        if (dev == NULL) {
		BUG();
		return -EIO;
        }

	/* Find the last value in the pre-stored response. */
        last = idx + 15;
        while ((last != idx) && (*last == 0)) last--;

	/* Send pre-stored response, set command status on the last value. */
        while (idx != last) {
		hil_intr_frame(port, dev, HIL_ERR_INT | /* devid | */ *idx);
		idx++;
        }
	hil_intr_frame(port, dev, HIL_ERR_INT | HIL_FRAME_CMD | 
		       /* devid | */ *idx);
	return 0;

 iserr:
	kbd->casc_wshift = 24;
	kbd->cascbuf = 0;
	return -ENOTSUPP; /* Normal HIL (MLC) driver returns -EIO on error */
}

static int hil_kbd_cascade_open(struct serio *port) {
	return 0;
}

void hil_kbd_cascade_close(struct serio *port) {
	return;
}


/* Linkage into the linux-input system. */

static void hil_kbd_disconnect(struct serio *serio)
{
	struct hil_kbd *kbd;

	kbd = (struct hil_kbd *)serio->private;
	if (kbd == NULL) {
		BUG();
		return;
	}

	input_unregister_device(&kbd->dev);
	serio_close(serio);
	kbd->serio = NULL;
	clear_bit(kbd->devphys[9] - '0', hil_kbd_idsused);

	/* Temporary hack until input/serio moves to new driver model */
	if (kbd->casc_need_reg == 2) {
		struct hil_kbd **slot;
		slot = hil_kbd_freeme;
		while (*slot) slot++;
		*slot = kbd;
	}
	else {
		kfree(kbd);
	}
}

static void hil_kbd_setup(struct hil_kbd *kbd, uint8_t id) 
{
	int i;
	kbd->dev.evbit[0]	= BIT(EV_KEY) | BIT(EV_REP);

	/* TODO: get from HIL "PROMPT" info */
	kbd->dev.ledbit[0]	= BIT(LED_NUML) | BIT(LED_CAPSL);
	kbd->dev.ledbit[0]	|= BIT(LED_SCROLLL);

	kbd->dev.keycodemax	= HIL_KEYCODES_SET1_TBLSIZE;
	kbd->dev.keycodesize	= sizeof(hil_kbd_set1[0]);
	kbd->dev.keycode	= hil_kbd_set1;

	kbd->dev.name = strlen((char *)kbd->rnm) ? 
		(char *)kbd->rnm : HIL_GENERIC_NAME;
	kbd->dev.phys		= kbd->devphys;
	kbd->dev.uniq		= kbd->devphys;
	memcpy(kbd->devphys, "hil/inputX", 11);
	kbd->devphys[9]		= '0' + id;

	kbd->dev.rep[REP_DELAY] = HZ/4; /* auto-repeat values */
	kbd->dev.rep[REP_PERIOD]= HZ/8;

	kbd->dev.id.bustype	= BUS_HIL;
	kbd->dev.id.vendor	= PCI_VENDOR_ID_HP;
	kbd->dev.id.product	= 0x0001; /* TODO: get from kbd->rsc */
	kbd->dev.id.version	= 0x0100; /* TODO: get from kbd->rsc */

	for (i = 0; i < 128; i++) {
		set_bit(hil_kbd_set1[i], kbd->dev.keybit);
		set_bit(hil_kbd_set3[i], kbd->dev.keybit);
	}
	clear_bit(0, kbd->dev.keybit);

	/* Initialize the cascaded serio in case we need it. */
	kbd->cascade.type 	= SERIO_HIL_MLC | SERIO_HIL;
	kbd->cascade.write 	= hil_kbd_cascade_write;
	kbd->cascade.open	= hil_kbd_cascade_open;
	kbd->cascade.close	= hil_kbd_cascade_close;
	kbd->cascade.name	= "cascaded device off HIL keyboard";
	kbd->cascade.phys	= kbd->cascphys;
	memcpy(kbd->cascphys, "hilkbd/seriox", 14);
	kbd->cascphys[12]	= '0' + id;

	kbd->casc_wshift = 24;
	kbd->cascbuf = 0;

	input_register_device(&kbd->dev);
	printk(KERN_INFO "input: %s (%s)\n", kbd->dev.phys, kbd->dev.name);
	set_bit(id, hil_kbd_idsused);
}

static void hil_kbd_connect(struct serio *serio, struct serio_dev *dev)
{
	struct hil_kbd	*kbd;
	uint8_t		id, did, *idd;

	if (serio->type != (SERIO_HIL_MLC | SERIO_HIL)) return;

	if (strncmp(serio->phys, "hil/serio", 9) &&
	    strncmp(serio->phys, "hilape/serio", 12)) return;

	if (!(kbd = kmalloc(sizeof(struct hil_kbd), GFP_KERNEL))) return;
	memset(kbd, 0, sizeof(struct hil_kbd));

	if (serio_open(serio, dev)) goto bail0;

	serio->private = kbd;
	kbd->serio = serio;
	init_input_dev(&kbd->dev);
	kbd->dev.private = kbd;
	init_MUTEX_LOCKED(&(kbd->sem));

	for (id = 0; id < 32; id++) {
		if (!test_bit(id, hil_kbd_idsused)) break;
	}
	if (id > 31) goto bail1;

	/* When in APE mode we get no device ID information, so we 
	 * just have to pretend detection succeeded.  We delay registering 
	 * the cascade serio until input received.  We cannot do so here 
	 * because we are inside an input-core traversal.
	 */
	if (!strncmp(serio->phys, "hilape/serio", 12)) {
		kbd->ape = 1;
		hil_kbd_setup(kbd, id);
		kbd->rsc[0] = HIL_CMD_RSC;
		kbd->rnm[0] = 'c';
		kbd->rnm[1] = 'a';
		kbd->rnm[2] = 's';
		kbd->rnm[3] = 'c';
		kbd->rnm[4] = HIL_CMD_RNM;
		kbd->exd[0] = HIL_CMD_EXD;
		/* IDD is created later based on input received. */
		kbd->casc_need_reg = 1;
		kbd->axdepth = 0;
		return;
	}

	/* Get device info.  MLC driver is guaranteed to respond. */
	hil_write_frame(serio, HIL_FRAME_CMD | HIL_CMD_IDD);
	down(&(kbd->sem));

	hil_write_frame(serio, HIL_FRAME_CMD | HIL_CMD_RSC);
	down(&(kbd->sem));

	hil_write_frame(serio, HIL_FRAME_CMD | HIL_CMD_RNM);
	down(&(kbd->sem));

	hil_write_frame(serio, HIL_FRAME_CMD | HIL_CMD_EXD);
	down(&(kbd->sem));

	up(&(kbd->sem));

	did = kbd->idd[0];
	idd = kbd->idd + 1;
	switch (did & HIL_IDD_DID_TYPE_MASK) {
	case HIL_IDD_DID_TYPE_KB_INTEGRAL:
	case HIL_IDD_DID_TYPE_KB_ITF:
	case HIL_IDD_DID_TYPE_KB_RSVD:
	case HIL_IDD_DID_TYPE_CHAR:
		printk(KERN_INFO PREFIX "HIL kbd (did = 0x%02x, lang = %s)\n",
		       did, 
		       hil_language[did & HIL_IDD_DID_TYPE_KB_LANG_MASK]);
		break;
	default:
		goto bail1;
	}
	kbd->ape = 0;
	hil_kbd_setup(kbd, id);

	/* Enable Keyswitch Autorepeat 1 */
	hil_write_frame(serio, HIL_FRAME_CMD | HIL_CMD_EK1); 
	down(&(kbd->sem));
	up(&(kbd->sem));

	/* Delay registering cascade serio until input received.  We
	 * cannot do so here inside an input-core traversal, at least
	 * until the new driver-model is implemented for linux-input.
	 */
	kbd->casc_need_reg = 0;
	if(HIL_IDD_NUM_BUTTONS(idd) || HIL_IDD_NUM_AXES_PER_SET(*idd)) {
		printk(KERN_INFO PREFIX "Combo device: Will do slave serio\n");
		kbd->casc_need_reg = 1;
		kbd->axdepth = 0;
	}

	return;
 bail1:
	serio_close(serio);
 bail0:
	kfree(kbd);
}

struct serio_dev hil_kbd_serio_dev = {
	.name =		"HIL protocol keyboard",
	.connect =	hil_kbd_connect,
	.disconnect =	hil_kbd_disconnect,
	.interrupt =	hil_kbd_interrupt
	/* TODO: write_wakeup, reconnect, cleanup */
};

static int __init hil_kbd_init(void)
{
	bitmap_zero(hil_kbd_idsused, 32);
	serio_register_device(&hil_kbd_serio_dev);
	printk(KERN_INFO "HIL keyboard support loaded.\n");
        return 0;
}
                
static void __exit hil_kbd_exit(void)
{
	struct hil_kbd **kbd;
	serio_unregister_device(&hil_kbd_serio_dev);

	/* Temporary hack until input/serio moves to new driver model */
	kbd = hil_kbd_freeme;
	while (kbd - hil_kbd_freeme < 32) {
		if (*kbd) {
			serio_unregister_port(&((*kbd)->cascade));
			kfree(*kbd);
		}
		kbd++;
	}

	printk(KERN_INFO "HIL keyboard support unloaded.\n");
}
                        
module_init(hil_kbd_init);
module_exit(hil_kbd_exit);
