/*
 * HP i8042 SDC + HIL Auto-Poll-Enabled ("simple") driver.
 *
 * Copyright (c) 2004 Brian S. Julin
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL").
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *
 * References:
 *
 * System Device Controller Microprocessor Firmware Theory of Operation
 *      for Part Number 1820-4784 Revision B.  Dwg No. A-1820-4784-2
 * HP-HIL Technical Reference Manual.  Hewlett Packard Product No. 45918A
 * Original hilkbd.c driver (Deller, Wilcox, Blundell)
 *
 * Description:
 *
 * This is the "simple" HIL driver which allows the HP i8042 SDC firmware
 * to take care of most of the details of managing the HIL MLC circuit.
 * In this manner it is very similar to the USB HIDBP driver.
 * 
 * There are many limitations to using this driver as opposed to the 
 * "Full HIL MLC support" driver (not in mainline kernel yet), however, 
 * for simple systems with only keyboards and mice which are seldom
 * detached, it is recommended over the complex driver.  For more 
 * information, see Documentation/input/hil.txt
 *
 */

#include <linux/hp_sdc.h>
#include <linux/hil.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/device.h>
#include <linux/serio.h>

#define PREFIX "HILAPE: "
#define HIL_GENERIC_NAME "HIL APE mode (read-only) driver"

MODULE_AUTHOR("Brian S. Julin <bri@abrij.org>");
MODULE_DESCRIPTION("HP i8042 SDC HIL APE Driver");
MODULE_LICENSE("Dual BSD/GPL");

EXPORT_SYMBOL(hil_bus_type);

static hp_sdc_irqhook hilape_isr;
static hp_sdc_func hilape_func;
struct tasklet_struct hilape_task;

/* Driver-wide state */
int do_rescan;					/* nimsaddr changed */
static unsigned char nimsaddr, nimsaddr_old;	/* which devs are keyboards */
uint8_t hilape_kbd_arr = 3;			/* global key repeat rate */
uint8_t hilape_kbd_ard = 25;			/* global key repeat delay */

/* Devices that appear on the HIL bus (and for now, serios for each.) */

static struct device hilape_devs[7];
static struct serio hilape_serios[7];
char hilape_serio_names[7 * 14];
hil_frame hilape_serio_cmd[7];
int hilape_serio_cmd_idx[7];
static int hilape_devs_good[7] = {1,1,1,1,1,1,1};

static void hilape_dev_release(struct device *dev) { return; }

/* Interrupt handler and globals for use by ISR only. */
static char ibuf[64];	/* raw status/data from the SDC stored here */
static int iidx;	/* current index into ibuf */
static int oidx;	/* marks end of a valid ibuf if nonzero */
static unsigned long sdc_irqs, hil_irqs, nmi_irqs; 
static unsigned long auto_conf, err, busy, timeout;
static unsigned long bad_seq;
static void hilape_isr (int irq, void *dev_id, uint8_t status, uint8_t data) 
{
	int i, cnt;

	/* Check for overflows */
	if (iidx > 30) {
		iidx = 0;
		bad_seq++;
		return;
	}
	if ((status & HP_SDC_STATUS_IRQMASK) == HP_SDC_STATUS_HILDATA) {
		/* This is a "6X" interrupt. */
		hil_irqs++;
		if (iidx < 2) goto store;

		/* Check for sequence termination */
		if (((ibuf[iidx - 2] & HP_SDC_STATUS_IRQMASK) 
		     == HP_SDC_STATUS_HILCMD) &&
		    (ibuf[iidx - 1] & (HIL_FRAME_CMD >> 8))) 
			goto check;
		goto store;

	}
	/* This is a "5X" interrupt.  Could be an error. */
	sdc_irqs++;
	if (!(data & HP_SDC_HIL_ISERR)) goto store;
	if (data == HP_SDC_HIL_RC_DONE) {
		auto_conf++;
		/* Locking is really not needed, 200ms between reconfigs. */
		do_rescan++;
		tasklet_schedule(&hilape_task);
	}
	if (data == HP_SDC_HIL_ERR) err++;
	if (data == HP_SDC_HIL_TO) timeout++;
	if (data == HP_SDC_HIL_RC) busy++;
	iidx = 0;
	return;
 store:
	/* Store data for later processing */
	ibuf[iidx++] = status;
	ibuf[iidx++] = data;
	return;
 check:
	/* Check sequence for integrity, and copy over if good. */
	if ((ibuf[0] & HP_SDC_STATUS_IRQMASK) != HP_SDC_STATUS_HILCMD)
		goto bad;
	for (i = 0, cnt = 0; i < iidx; i += 2) {
		if ((ibuf[i] & HP_SDC_STATUS_IRQMASK) == HP_SDC_STATUS_HILDATA)
			cnt++;
	}
	if (cnt < 1) goto bad;
	if ((data & HIL_CMDCT_POL) != cnt) goto bad;
	ibuf[iidx++] = status;
	ibuf[iidx++] = data;

	/* Locking is really not needed, 20ms between polls. */

	oidx = iidx;
	tasklet_schedule(&hilape_task);
	iidx = 0;
	return;
 bad:
	bad_seq++;
	iidx = 0;
	return;
}

/* Define a bus for HIL devices */
static int hil_bus_match(struct device *dev, struct device_driver *drv) {

	/* We only allow hil_kbd driver to attach directly -- it will
	 * spawn a hil_ptr subdevice if needed.  Actually we should
	 * also check nimsaddr here, but another day for that.
	 */
	if (!strncmp(drv->name, "hil_kbd", 7))
		return 0;
	return 1;
}

struct bus_type hil_bus_type = {
	.name   = "hil",
	.match  = hil_bus_match,
};

static int hilape_probe (struct device * dev);
static int hilape_remove (struct device *dev);

static hp_sdc_func hilape_func = {
	.driver = {
		.name   = "hilape",
		.bus    = &hp_sdc_bus_type,
		.probe  = hilape_probe,
		.remove = hilape_remove,
	},
	.did            = '5', /* HP_SDC_CMD_DO_HIL */
	.irq            = HP_SDC_STATUS_HILCMD,
	.callback       = hilape_isr,
};


/* Bus statistics.  Technically we should lock these since 
 * ISR increments, but they are only really for debugging use.
 */
static ssize_t hilape_show_sdc_irqs(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",sdc_irqs);
}
static BUS_ATTR(sdc_irqs,0444,hilape_show_sdc_irqs,NULL);

static ssize_t hilape_show_hil_irqs(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",hil_irqs);
}
static BUS_ATTR(hil_irqs,0444,hilape_show_hil_irqs,NULL);

static ssize_t hilape_show_nmi_irqs(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",nmi_irqs);
}
static BUS_ATTR(nmi_irqs,0444,hilape_show_nmi_irqs,NULL);

static ssize_t hilape_show_auto_conf(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",auto_conf);
}
static BUS_ATTR(auto_conf,0444,hilape_show_auto_conf,NULL);

static ssize_t hilape_show_err(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",err);
}
static BUS_ATTR(err,0444,hilape_show_err,NULL);

static ssize_t hilape_show_busy(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",busy);
}
static BUS_ATTR(busy,0444,hilape_show_busy,NULL);

static ssize_t hilape_show_timeout(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",timeout);
}
static BUS_ATTR(timeout,0444,hilape_show_timeout,NULL);

static ssize_t hilape_show_bad_seq(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%lu\n",bad_seq);
}
static BUS_ATTR(bad_seq,0444,hilape_show_bad_seq,NULL);

static int hilape_sdccmd(uint8_t cmd, uint8_t arg) {
	struct semaphore tsem;
	hp_sdc_transaction t;
	uint8_t tseq[4];
	int i = 0;

	tseq[i++] = 
		HP_SDC_ACT_PRECMD | HP_SDC_ACT_DATAOUT | HP_SDC_ACT_SEMAPHORE;
	tseq[i++] = cmd;
	tseq[i++] = 1;
	tseq[i++] = arg;
	t.endidx = i;
	t.seq = tseq;
	t.act.semaphore = &tsem;
	init_MUTEX_LOCKED(&tsem);
        if (hp_sdc_enqueue_transaction(&t)) 
		return -ENODEV;
	down_interruptible(&tsem);  /* Put ourselves to sleep for results. */
	if (tseq[i-1] == HP_SDC_ACT_DEAD) return -ENODEV;
	return 0;
}

static ssize_t hilape_do_reconf(struct bus_type *bt, const char *b, size_t c) 
{
	ssize_t ret;
	ret = hilape_sdccmd(HP_SDC_CMD_SET_LPC, 
			    HP_SDC_LPC_APE_IPF | HP_SDC_LPC_RC);
	if(ret) return ret;
	return c;
}
static BUS_ATTR(do_reconf,0220,NULL,hilape_do_reconf);

static ssize_t hilape_show_kbd_ard(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%u\n",hilape_kbd_ard);
}
static ssize_t hilape_set_kbd_ard(struct bus_type *bt, const char *b, size_t c)
{
	size_t ret;

	ret = c;
	if (ret > 3) ret = 3;
	hilape_kbd_ard = 0;
	while(ret--) {
		if (*b > '9' || *b < '0') break;
		hilape_kbd_ard *= 10;
		hilape_kbd_ard += *b - '0';
		b++;
	}
	ret = hilape_sdccmd(HP_SDC_CMD_SET_ARD, hilape_kbd_ard);
	if (ret) return ret;
	return c;
}
static BUS_ATTR(kbd_ard,0664,hilape_show_kbd_ard,hilape_set_kbd_ard);

static ssize_t hilape_show_kbd_arr(struct bus_type *bt, char * buf) {
	return snprintf(buf,22,"%u\n",hilape_kbd_arr);
}
static ssize_t hilape_set_kbd_arr(struct bus_type *bt, const char *b, size_t c)
{
	size_t ret;

	ret = c;
	if (ret > 3) ret = 3;
	hilape_kbd_arr = 0;
	while(ret--) {
		if (*b > '9' || *b < '0') break;
		hilape_kbd_arr *= 10;
		hilape_kbd_arr += *b - '0';
		b++;
	}
	ret = hilape_sdccmd(HP_SDC_CMD_SET_ARR, hilape_kbd_arr);
	if (ret) return ret;
	return c;
}
static BUS_ATTR(kbd_arr,0664,hilape_show_kbd_arr,hilape_set_kbd_arr);

struct bus_attribute *hil_attrs[12] = {
	&bus_attr_sdc_irqs,
	&bus_attr_hil_irqs,
	&bus_attr_nmi_irqs,
	&bus_attr_auto_conf,
	&bus_attr_err,
	&bus_attr_busy,
	&bus_attr_timeout,
	&bus_attr_bad_seq,
	&bus_attr_do_reconf,
	&bus_attr_kbd_ard,
	&bus_attr_kbd_arr,
	NULL
};

static int hilape_probe (struct device * dev) { return 0; }

static int hilape_remove (struct device *dev) {
	hp_sdc_unhook_irq(&hilape_func);
	return 0;
}

/* Facilities to rescan the bus when an auto-configure is received. */
static uint8_t hilape_rescan_tseq[4] = { 
	HP_SDC_ACT_PRECMD | HP_SDC_ACT_DATAIN | HP_SDC_ACT_CALLBACK,
	HP_SDC_CMD_READ_KBN,
	1,
};

static void hilape_rescan(int x, void *xx, uint8_t xxx, uint8_t xxxx) {
	int i;

	nimsaddr_old = nimsaddr;
	nimsaddr = hilape_rescan_tseq[3];

	if (do_rescan > 1) { /* Module was just loaded */
		for (i = 0; i < 7; i++) {
			hilape_devs_good[i] = device_register(hilape_devs + i);
			/* No error checking here, since serio code in flux */
			serio_register_port(hilape_serios + i);
			printk(KERN_INFO "serio: %s: %s at HIL ID #%c\n",
			       hilape_serios[i].phys,
			       hilape_serios[i].name,
			       hilape_serios[i].phys[12] + 1);
		}
	}
	else {
		for (i = 0; i < 7; i++) {
			if ((nimsaddr ^ nimsaddr_old) & (1 << i)) {
				if (!hilape_devs_good[i]) {
					hilape_devs_good[i] = -1;
					serio_unregister_port(hilape_serios+i);
					device_unregister(hilape_devs + i);
				}
				hilape_devs_good[i] = 
					device_register(hilape_devs + i);
				/* No error checking here serio code in flux */
				serio_register_port(hilape_serios + i);
				printk(KERN_INFO "serio: %s reregistered\n",
				       hilape_serios[i].phys);
			}
		}
	}
	do_rescan = 0;
}

static hp_sdc_transaction hilape_rescan_act = {
	.endidx = 4,
	.seq = hilape_rescan_tseq,
	.act = {
		.irqhook = hilape_rescan,
	},
};

/* Timer task (bottom half) */
static void hilape_do_task(unsigned long foo) {
	int widx, id = 0, cnt = 0;
	struct serio *s = NULL;

	/* do_rescan is only run after autoreconfigs, which take 
	 * a long time to complete.  So we should be safe to 
	 * check/reset it even though the ISR writes to it.
	 */
	if (do_rescan) {
		hp_sdc_enqueue_transaction(&hilape_rescan_act);
		return;
	}

	/* Because autopolls are limited to 20ms, we are pretty
	 * much guaranteed to run well before the next one.  So
	 * we do not need to lock the buffers/indices.
	 */
	id = ibuf[1] & HP_SDC_HIL_R1MASK;
	if (ibuf[1] & (HIL_FRAME_CMD >> 8)) goto isbug;
	s = hilape_serios + id - 1;

	/* We send the command response, skipping the initial
	 * SDC-generated command.  At device-chaining markers,
	 * we send a packet terminator instead.  We do nothing 
	 * with the final SDC-generated post-terminator data 
	 * count item.
	 */
	for (widx = 2; widx < oidx; widx += 2) {
		uint8_t status, data;

		status = ibuf[widx];
		data = ibuf[widx+1];

		if ((status & HP_SDC_STATUS_IRQMASK) == HP_SDC_STATUS_HILDATA)
			goto isdata;
		if ((status & HP_SDC_STATUS_IRQMASK) != HP_SDC_STATUS_HILCMD)
			goto isbug;

		if (!id) goto skip;
		s = hilape_serios + id - 1;
		if (s->dev) {
			s->dev->interrupt(s, 0, 0, NULL);
			s->dev->interrupt(s, HIL_ERR_INT>>16, 0, NULL);
			s->dev->interrupt(s, id | (HIL_FRAME_CMD>>8),0, NULL);
			s->dev->interrupt(s, HIL_CMD_POL | cnt, 0, NULL);
		}
	skip:
		id = data & HP_SDC_HIL_R1MASK;
		if (!id || (data & (HIL_FRAME_CMD >> 8))) {
			if (widx + 4 != oidx) goto isbug;
			break;
		}
		if (!id) goto isbug;
		cnt = 0;
		continue;
	isdata:
		if (!id) goto isbug;
		if (s->dev) { 
			s->dev->interrupt(s, 0, 0, NULL);
			s->dev->interrupt(s, HIL_ERR_INT >> 16, 0, NULL);
			s->dev->interrupt(s, id, 0, NULL);
			s->dev->interrupt(s, data, 0, NULL);  
		}
		cnt++;
	}
	oidx = 0;
	return;

 isbug:
	printk(KERN_WARNING PREFIX "HIL APE demultiplex error\n");
	oidx = 0;
	return;
}

/* Module insertion and removal */

static int __init hilape_hwinit(void) {
	struct semaphore tsem;
	hp_sdc_transaction t;
	uint8_t tseq[12];
	int i = 0;

	tseq[i++] = HP_SDC_ACT_PRECMD | HP_SDC_ACT_DATAOUT;
	tseq[i++] = HP_SDC_CMD_SET_ARR;
	tseq[i++] = 1;
	tseq[i++] = hilape_kbd_arr;
	tseq[i++] = HP_SDC_ACT_PRECMD | HP_SDC_ACT_DATAOUT;
	tseq[i++] = HP_SDC_CMD_SET_ARD;
	tseq[i++] = 1;
	tseq[i++] = hilape_kbd_ard;
	tseq[i++] = 
		HP_SDC_ACT_PRECMD | HP_SDC_ACT_DATAOUT | HP_SDC_ACT_SEMAPHORE;
	tseq[i++] = HP_SDC_CMD_SET_LPC;
	tseq[i++] = 1;
	tseq[i++] = HP_SDC_LPC_APE_IPF | HP_SDC_LPC_RC;
	t.endidx = i;
	t.seq = tseq;
	t.act.semaphore = &tsem;
	init_MUTEX_LOCKED(&tsem);
        if (hp_sdc_enqueue_transaction(&t)) 
		return -ENODEV;
	down_interruptible(&tsem);  /* Put ourselves to sleep for results. */
	if (tseq[i-1] == HP_SDC_ACT_DEAD) return -ENODEV;
	return 0;
}

/* Pre-device-class serio device support.  This will change signifigantly
 * when serios graduate to the new driver-model.  For now we just
 * instantiate them alongside the driver-model nodes.
 */

static int hilape_serio_write(struct serio *port, unsigned char c) {
	return -ENOTSUPP; /* Normal HIL driver returns -EIO on error */
}

static int hilape_serio_open(struct serio *port) {
	return 0;
}

void hilape_serio_close(struct serio *port) {
	return;
}



static int __init hilape_init(void)
{
	int ret, i;

	iidx = oidx = 0;
	sdc_irqs = hil_irqs = nmi_irqs = 0;
	auto_conf = err = busy = timeout = 0;
	bad_seq = 0;
	do_rescan = 2;
	nimsaddr = nimsaddr_old = 0;

	ret = driver_register(&hilape_func.driver);
	if (ret) return ret;
	ret = bus_register(&hil_bus_type);
	if (ret) goto bail0;

	i = 0;
	while (hil_attrs[i] != NULL) {
		ret = bus_create_file(&hil_bus_type, hil_attrs[i]);
		if (ret) break;
		i++;
	}
	if (hil_attrs[i] != NULL) goto bail1;

	for(i = 0; i < 7; i++) {
		char id;

		id = '1' + i;
		hilape_devs[i].kobj.name[0] = id;
		hilape_devs[i].kobj.name[1] = 0;
		hilape_devs[i].bus_id[0] = id;
		hilape_devs[i].bus_id[1] = 0;
		hilape_devs[i].bus = &hil_bus_type;
		hilape_devs[i].release = hilape_dev_release;

		hilape_serios[i].type = SERIO_HIL_MLC | SERIO_HIL;
		hilape_serios[i].write = hilape_serio_write;
		hilape_serios[i].open = hilape_serio_open;
		hilape_serios[i].close = hilape_serio_close;
		hilape_serios[i].name = "HIL APE mode protocol";
		hilape_serios[i].phys = hilape_serio_names + i * 14;
		memcpy(hilape_serios[i].phys, "hilape/seriox", 14);
		hilape_serios[i].phys[12] = id - 1;

		hilape_serio_cmd_idx[i] = 0;
	}

        /* Create the bottom half */
	tasklet_init(&hilape_task, hilape_do_task, 0);

	/* Set up the i8042 for autopolling but no keyboard cooking */
	ret = hilape_hwinit();
	if (ret) goto bail2;

	printk(KERN_INFO "HIL APE (SDC) mode support loaded.\n");
	return 0;

 bail2:
	tasklet_kill(&hilape_task);
 bail1:
	while (--i >= 0) {
		bus_remove_file(&hil_bus_type, hil_attrs[i]);
	}
	bus_unregister(&hil_bus_type);
 bail0:
	driver_unregister(&hilape_func.driver);
	return ret;
}

static void __exit hilape_exit(void)
{
	int i;

	tasklet_kill(&hilape_task);
	for(i = 0; i < 7; i++) {
		if (!hilape_devs_good[i]) {
			device_unregister(hilape_devs + i);
			serio_unregister_port(hilape_serios + i);
		}
	}
	i = 0;
	while (hil_attrs[i]) {
		bus_remove_file(&hil_bus_type, hil_attrs[i]);
		i++;
	}
	bus_unregister(&hil_bus_type);
	driver_unregister(&hilape_func.driver);
	printk(KERN_INFO "HIL APE (SDC) mode support unloaded.\n");
}

module_init(hilape_init);
module_exit(hilape_exit);
